-- this file is for xc9572xl-7tq100 chip which controls an 8 Mbit FLASH memory
-- for DCC board FPGA configuration. ( It may be used to store TTCrx configuration
-- data too )
-- FLASH space assignment:
-- 0x40000 - 0x4E750	FLASH access configuration data for PCI3
-- 0x50000 - 0x5E750	normal configuration data for PCI1
-- 0x60000 - 0x6E750	configuration data for PCI2
-- 0x70000 - 0x7E750	configuration data for PCI3
-- 0x80000 - 		configuration data for xc2v1000
-- after power on reset, configuration starts from 0x50000. If PCI3 configuration
-- is successful, the process continues with PCI2, then PCI1 and finally xc2v1000.
-- the maximum clk speed is 32MHz, limited by the acex spec. The actual configure
-- clock frequency is 16MHz
-- Either normal reconfiguration or FALSH access reconfiguration could be started
-- by software.
-- To start a reconfiguration, write the starting address: 0x4F100 for normal reconfiguration
-- and 0x3F100 for FLASH access reconfiguration. Then write 0xA5 to data register.
-- And finally, set sp_i(5) to '1'.
-- 
-- Access of the FLASH memory itself is under PCI3 control. Address must be set by
-- the software and does not change by itself. Data write of PCI3 starts the FLASH
-- command execution. For FLASH commands, please refer to AMD AM29F080b data sheet.
-- Avoid using chip erase or erasing sector SA4 and SA5 at the same time. This may
-- lead to the situation where you have to use JTAG to configure the PCI3 chip.
-- 
-- address register is shifted in from sp_i(4) only when sp_i(0) is '1'
--
-- data register is loaded from mdq input when sp_i(1) = '1' or status signals when
-- sp_i(2) = '0' and sp_i(3) = '0', otherwise it keeps on shifting from sp_i(4)
-- 
-- sp_i and sp_o signal definition
-- sp_i(0) load address flag
-- sp_i(1) read FLASH flag 
-- sp_i(2) write FLASH flag
-- sp_i(3) shift data register flag
-- sp_i(4) address/data serial input
-- sp_i(5) reconfiguration flag
-- sp_i(6) I2C bus clock signal scl
-- sp_i(7) I2C bus data input
--
-- sp_o(0) is data register bit 7
-- sp_o(1) is conf_clk
-- sp_o(2) is I2C bus data output
-- sp_o(3) is virtex2 done signal

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_arith.all;

entity dcc_conf is
    port (
-- configure clock, coming from local control signal L_CTL19, 32MHz
        clk: in STD_LOGIC;
-- power on reset, active low
        porn: in STD_LOGIC;
-- spare signals to and from pci3 interface chip
        sp_i: in STD_LOGIC_VECTOR (8 downto 0);
        sp_o: out STD_LOGIC_VECTOR (3 downto 0);
-- flash memory signals
        ma: out STD_LOGIC_VECTOR (19 downto 0);
        mdq: inout STD_LOGIC_VECTOR (7 downto 0);
        mrstn: out STD_LOGIC;
        mwen: out STD_LOGIC;
        mcen: out STD_LOGIC;
        moen: out STD_LOGIC;
-- reset outputs, low active
        rstn: out STD_LOGIC_VECTOR (3 downto 1);
-- virtex chip configure pins
        vtx_done: in std_logic;
        vtx_init: in std_logic;
        vtx_clk: out std_logic;
        vtx_data: out std_logic;
        vtx_confn: out std_logic;
-- pci1 chip configure pins
        pci1_done: in std_logic;
        pci1_dclk: out std_logic;
        pci1_data: out std_logic;
        pci1_confn: out std_logic;
-- pci2 chip configure pins
        pci2_done: in std_logic;
        pci2_dclk: out std_logic;
        pci2_data: out std_logic;
        pci2_confn: out std_logic;
-- pci3 chip configure pins
        pci3_idone: in std_logic;
        pci3_done: in std_logic;
        pci3_dclk: out std_logic;
        pci3_data: out std_logic;
        pci3_confn: out std_logic;
        dcc_clk: out std_logic;
-- virtex chip power down control
        pwrdownn: out std_logic;
-- TTCrx I2C connection
        sda: inout std_logic;
        scl: out std_logic
    );
end dcc_conf;

architecture dcc_conf_arch of dcc_conf is
signal vcc, gnd : std_logic;
signal powerondelay, en_conf, re_conf, conf_clk, en_dclk, conf_fail, mdq2data: std_logic;
signal load_addr, ad_in, shift_data, shift_data_in: std_logic;
signal read_flash, write_flash, flash_we_q, flash_we, shift_addr, backup: std_logic;
signal div8:	integer range 0 to 7;
signal data: std_logic_vector(7 downto 0);
signal addr: std_logic_vector(19 downto 0);
attribute INIT: string;
attribute INIT of powerondelay, en_conf: signal is "S";
begin
-- define vcc and gnd
vcc <= '1';
gnd <= '0';
-- power on reset fanout
rstn(1) <= porn;
rstn(2) <= porn;
rstn(3) <= porn;
mrstn <= porn;
dcc_clk <= clk;
-- no power down
pwrdownn <= vcc;
-- copy addr to ma
G1: for I in ma'range generate
  ma(I) <= addr(I);
end generate G1;
-- memory data and control signals
-- default to read from memory
mdq <= data when flash_we_q = '1' else (others => 'Z');
-- mcen is used to control FLASH write
moen <= write_flash and pci3_done and not en_conf;
mcen <= gnd;
mwen <= not (write_flash and not shift_data and pci3_done and not en_conf);
-- interface to pci3 chip
load_addr <= sp_i(0);
read_flash <= sp_i(1);
write_flash <= sp_i(2);
shift_data_in <= sp_i(3);
ad_in <= sp_i(4);
-- write should be high for at least 50ns + output delay + 1 clk cycle, i.e. 2 conf_clk
flash_we <= write_flash and not shift_data;
scl <= sp_i(6);
sda <= '0' when (sp_i(7) = '0') else 'Z';
-- conf_clk sent to pci3 interface
sp_o(0) <= data(7);
sp_o(1) <= not conf_clk;
sp_o(2) <= sda;
sp_o(3) <= vtx_done;
process(clk,porn)
begin
  if(porn = '0')then
    powerondelay <= '1';
    backup <= '0';
-- enable config
    en_conf <= '1';
-- set addr to 0x4F100
    addr <= x"4F100";
    data <= x"00";
    en_dclk <= '0';
    div8 <= 0;
    pci1_confn <= '1';
    pci2_confn <= '1';
    pci3_confn <= '1';
    vtx_confn <= '1';
    pci1_dclk <= '0';
    pci2_dclk <= '0';
    pci3_dclk <= '0';
    vtx_clk <= '1';
    conf_clk <= '0';
    conf_fail <= '0';
    mdq2data <= '0';
    shift_data <= '0';
    flash_we_q <= '0';
  elsif(clk'event and clk = '1')then
    flash_we_q <= not shift_data and write_flash and pci3_done and not en_conf;
-- config_n signals
    if(addr(19 downto 8) = x"4F1" and en_conf = '1')then
      pci1_confn <= powerondelay;
    else
      pci1_confn <= '1';
    end if;
    if(addr(19 downto 8) = x"5F1" and en_conf = '1')then
      pci2_confn <= powerondelay;
    else
      pci2_confn <= '1';
    end if;
-- due to DCM reset, virtex2 and pci3 must be reset at the same time
    if((addr(19 downto 8) = x"6F1" or addr(19 downto 8) = x"3F1") and en_conf = '1')then
      vtx_confn <= powerondelay;
      pci3_confn <= powerondelay;
    else
      vtx_confn <= '1';
      pci3_confn <= '1';
    end if;
-- configuration clock is clk divided by two and is 16MHz
    conf_clk <= not conf_clk;
-- configuration data
    vtx_data <= data(0) and addr(19) and en_dclk;
    vtx_clk <= conf_clk or not addr(19) or not en_dclk;
    if(addr(19 downto 16) = x"5" and en_dclk = '1')then
      pci1_data <= data(0);
      pci1_dclk <= conf_clk;
    else
      pci1_data <= '0';
      pci1_dclk <= '0';
    end if;
    if(addr(19 downto 16) = x"6" and en_dclk = '1')then
      pci2_data <= data(0);
      pci2_dclk <= conf_clk;
    else
      pci2_data <= '0';
      pci2_dclk <= '0';
    end if;
    if((addr(19 downto 16) = x"4" or addr(19 downto 16) = x"7") and en_dclk = '1')then
      pci3_data <= data(0);
      pci3_dclk <= conf_clk;
    else
      pci3_data <= '0';
      pci3_dclk <= '0';
    end if;
    if(conf_clk = '1')then
-- reset power on delay at 0x4F000
      if(addr(19 downto 8) = x"4F0")then
        powerondelay <= '0';
      end if;
-- en_dclk
      if(en_conf = '1' and addr(15 downto 0) = 0 and div8 = 7)then
        en_dclk <= not powerondelay;
      elsif((addr(19) = '0' or addr(18 downto 16) = 7) and addr(15 downto 12) = 15)then
        en_dclk <= '0';
      end if;
-- after virtex2 chip or both PCI3 has been tried, stop configuration
      if(powerondelay = '0' and en_conf = '1' and en_dclk = '0' and ( 
        (addr(19 downto 16) = x"F" and pci3_done = '1') or 
        (addr(19 downto 12) = x"4F" and (conf_fail = '1' or (pci3_done = '1' and re_conf = '0')))
        ))then
        en_conf <= '0';
        backup <= not addr(19);
      end if;
-- reconfigure command 
      if(sp_i(5) = '1' and flash_we_q = '1' and data = x"A5")then
        conf_fail <= '0';
        en_conf <= '1';
        re_conf <= '1';
      end if;
      if(re_conf = '1')then
        re_conf <= not en_dclk;
      end if;
-- address counter
-- automatic reconfigure if normal PCI3 configuration failed.
      if(pci3_done = '0' and addr(19) = '1')then
        conf_fail <= not powerondelay;
      end if;
      shift_addr <= not en_conf and pci3_done and load_addr;
-- load address from PCI3
      if( shift_addr = '1')then
        addr(0) <= ad_in;
        for i in 1 to 19 loop
          addr(i) <= addr(i-1);
        end loop;
-- during configuration, increment when div8 = 7
      elsif(div8 = 7)then
	addr <= addr + 1;
      end if;
-- data register
      shift_data <= shift_data_in;
      if(div8 = 6 or (read_flash = '1' and en_conf = '0' and pci3_done = '1'))then
        mdq2data <= not powerondelay;
      else
        mdq2data <= '0';
      end if;
      if(mdq2data = '1')then
        data <= mdq;
      elsif(en_conf = '1')then
        data(7) <= '0';
        for i in 1 to 7 loop
          data(i-1) <= data(i);
        end loop;
      elsif(shift_data = '1')then
        data(0) <= ad_in;
        for i in 1 to 7 loop
          data(i) <= data(i-1);
        end loop;
      elsif(write_flash = '0')then
        data (7) <= backup;
        data (6) <= not pci3_done;
        data (5) <= not pci2_done;
        data (4) <= not pci1_done;
        data (3) <= not vtx_done;
        data (2) <= not vtx_init;
        data (1) <= '0';
        data (0) <= '0';
      end if;  
-- configuration clock divider
      if(en_conf = '0')then
        div8 <= 0;
      else
        div8 <= div8 + 1;
      end if;
    end if;
  end if; 
end process;
end dcc_conf_arch;
